#ifndef BOARD_H
#define BOARD_H

#include <vector>
#include <list>
#include <stdexcept>

#include "square.h"
#include "sequence.h"
#include "geometry.h"
#include "actions.h"


namespace sudoku
{
  class Board
  {
  public:
    static const int UNKNOWN_RATING = -1;

    typedef Geometry::size_type    size_type;
    typedef std::list<Square>      SquareCollection;
    typedef std::vector<Sequence>  SequenceCollection;

    typedef Action<Square>    SquareAction;
    typedef Action<Sequence>  SequenceAction;

    class Exporter
    {
    private:
      const Board& m_board;

    public:
      explicit Exporter(const Board &b) : m_board(b) {}
      const Board& getBoard() const { return m_board; }
    };

    class geometry_violation : public std::runtime_error
    {
    public:
      geometry_violation(const std::string &msg) : runtime_error(msg) {}
    };

  private:
    Geometry m_geometry;
    SquareCollection m_squares;
    SequenceCollection m_sequences;

    int m_rating;
    std::string m_comment;

    template<class T>
    class Push : public Action<T>
    {
    public:
      void operator()(T &x) const { x.push(); }
    };

    template<class T>
    class SetTag : public std::unary_function<T, void>
    {
    private:
      unsigned *const m_array;
      size_t i;

    public:
      SetTag(unsigned *a) : m_array(a), i(0) {}
      void operator()(T &x) { x.setTag(m_array[i++]); }
    };

    class ByTag
    {
    public:
      bool operator()(const Square &s1, const Square &s2) const {
        return (s1.getTag() < s2.getTag());
      }
    };

    void createTopology();
    void copyNumbers(const SquareCollection&);

    struct Difficulty
    {
      int leastScore;
      const char *name;
    };

    static const Difficulty RATING[];

  public:
    const Geometry& getGeometry() const { return m_geometry; }

    size_type getMaxNumber() const { return getGeometry().getMaxNumber(); }
    size_type getGridWidth() const { return getGeometry().getGridWidth(); }
    size_type getGridHeight() const { return getGeometry().getGridHeight(); }

    size_type getGridsPerRow() const {
      return getGeometry().getGridsPerRow();
    }

    size_type getGridsPerColumn() const {
      return getGeometry().getGridsPerColumn();
    }

    bitmask_t getNumberMask() const { return getGeometry().getNumberMask(); }

    int getRating() const { return m_rating; }
    void setRating(int r) { m_rating = r; }

    const std::string& getComment() const { return m_comment; }
    void setComment(const std::string &c) { m_comment = c; }

    const SquareCollection& getSquares() const { return m_squares; }
    SquareCollection& getSquares() { return m_squares; }

    const SequenceCollection& getSequences() const { return m_sequences; }
    SequenceCollection& getSequences() { return m_sequences; }

    const Sequence& getSequence(unsigned number) const
                        { return m_sequences[number]; }
    Sequence& getSequence(unsigned number) { return m_sequences[number]; }

    Sequence& getRow(unsigned number) { return getSequence(number); }
    const Sequence& getRow(unsigned number) const
                        { return getSequence(number); }

    Sequence& getColumn(unsigned number)
                  { return getSequence(getMaxNumber() + number); }
    const Sequence& getColumn(unsigned number) const
                  { return getSequence(getMaxNumber() + number); }

    Sequence& getGrid(unsigned number)
                  { return getSequence(2*getMaxNumber() + number); }
    const Sequence& getGrid(unsigned number) const
                  { return getSequence(2*getMaxNumber() + number); }

    Sequence& getGrid(unsigned r, unsigned c)
                  { return getGrid(r*getGridsPerRow() + c); }
    const Sequence& getGrid(unsigned r, unsigned c) const
                  { return getGrid(r*getGridsPerRow() + c); }

    Sequence& getGridOf(unsigned row, unsigned column)
                { return getGrid( getGeometry().getGridNumber(row, column) ); }
    const Sequence& getGridOf(unsigned row, unsigned column) const
                { return getGrid( getGeometry().getGridNumber(row, column) ); }

    const Sequence& operator[](unsigned i) const { return getRow(i); }
    Sequence& operator[](unsigned i) { return getRow(i); }

    void doAllSequences(const SequenceAction&);
    void doAllSquares(const SquareAction&);

    void pushSequences();
    void pushSquares();
    void push();

    bool popSequences();
    bool popSquares();
    bool pop();

    void dropStack();
    void clear();

    const Square& find(const Square &s) const {
        return getRow(s.getRowIndex())[s.getColumnIndex()];
    }
    Square& find(const Square &s) {
        return getRow(s.getRowIndex())[s.getColumnIndex()];
    }
    Square& find(unsigned i) {
        const Geometry& g = getGeometry();
        return getRow(g.getRowIndex(i))[g.getColumnIndex(i)];
    }

    explicit Board(const Geometry&);
    Board(const Board&);
    ~Board();

    Board& operator=(const Board&);

    void dump(std::ostream&, const std::string&) const;

    static const char* getRating(int r);
    static const char* getRatingByIndex(int r);
    static int getRatingIndex(int r);

    static void randomize(SquareCollection&);
    static const std::string makeGridName(const Geometry&, unsigned i);
  };

  Board& Randomize(Board&);
  Board& RandomInit(Board&);
}

namespace std
{
  istream& operator>>(istream&, sudoku::Board&);
  ostream& operator<<(ostream&, const sudoku::Board&);
  ostream& operator<<(ostream&, const sudoku::Board::Exporter&);
}

#endif
