#include "canvas.h"
#include "board.h"
#include "square.h"
#include "window.h"
#include "document.h"

extern "C"
{
  #include "interface.h"
  #include "support.h"
}

namespace sudoku
{

const GdkColor Canvas::RED   = { 0, 65535, 0, 0 };
const GdkColor Canvas::BLUE  = { 0, 0, 0, 65535 };
const GdkColor Canvas::GREEN = { 0, 56*256, 134*256, 11*256 };


/*********************************************************************
 */
class DrawRef : public Document::SquareRefAction
{
private:
  Canvas *m_canvas;
  const GdkColor *m_colour;

public:
  DrawRef(Canvas *cv, const GdkColor *cl) : m_canvas(cv), m_colour(cl) {}
  void operator()(Square *s) const { m_canvas->draw(*s, m_colour); }
};


/*********************************************************************
 */
gboolean
on_canvas_configure(GtkWidget*, GdkEventConfigure*, gpointer user_data)
{
  sudoku::Canvas *canvas = static_cast<sudoku::Canvas*>(user_data);
  if (canvas->isDirty())
    canvas->repaint();
  return TRUE;
}


gboolean
on_canvas_expose_event(GtkWidget      *widget,
                       GdkEventExpose *event,
                       gpointer        user_data)
{
  sudoku::Canvas *canvas = static_cast<sudoku::Canvas*>(user_data);

  if (canvas->m_backingStore != NULL)
  {
    gdk_draw_drawable(widget->window,
	  	                widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
	                    canvas->m_backingStore,
		                  event->area.x, event->area.y,
		                  event->area.x, event->area.y,
		                  event->area.width,
                      event->area.height);
  }

  return FALSE;
}


gboolean
on_canvas_button_press_event(GtkWidget*,
                             GdkEventButton *event,
                             gpointer        user_data)
{
  sudoku::Canvas *canvas = static_cast<sudoku::Canvas*>(user_data);

  gint xx = sudoku::Canvas::getClickableDistance(event->x);
  gint yy = sudoku::Canvas::getClickableDistance(event->y);

  if ( (xx >= 0) &&
       (xx < canvas->getClickableWidth()) &&
       (yy >= 0) &&
       (yy < canvas->getClickableHeight()) )
  {
    Board* board = canvas->getBoard();
    const Geometry& geometry = board->getGeometry();

    gint row = sudoku::Canvas::getCellRowIndex(yy, geometry);
    gint col = sudoku::Canvas::getCellColumnIndex(xx, geometry);

    canvas->getMainWindow()->onClicked( (*board)[row][col] );
  }

  return FALSE;
}


/*********************************************************************
 */
Canvas::Canvas(Window *w)
  : Component( w, gtk_drawing_area_new() )
  , m_backingStore(NULL)
  , m_isDirty(false)
{
  GtkWidget *alignment;

  alignment = lookup_widget( GTK_WIDGET(w->getGtkWidget()),
                             "canvas_alignment" );
  if ( alignment != NULL )
  {
    gtk_container_add(GTK_CONTAINER(alignment), getGtkWidget());
    gtk_widget_set_events(getGtkWidget(),
                          GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK);

    g_signal_connect(static_cast<gpointer>(getGtkWidget()),
                     "configure_event",
                     G_CALLBACK(on_canvas_configure),
                     this);
    g_signal_connect(static_cast<gpointer>(getGtkWidget()),
                     "expose_event",
                     G_CALLBACK(on_canvas_expose_event),
                     this);
    g_signal_connect(static_cast<gpointer>(getGtkWidget()),
                     "button_press_event",
                     G_CALLBACK(on_canvas_button_press_event),
                     this);
    resize();
    show();
  }
}


/*********************************************************************
 */
Canvas::~Canvas()
{
  if (m_backingStore != NULL)
    g_object_unref(m_backingStore);
}


/*******************************************************************
 */
Document*
Canvas::getDocument() const
{
  return getMainWindow()->getDocument();
}


/*******************************************************************
 */
Board*
Canvas::getBoard() const
{
  return getDocument()->getBoard();
}


/*********************************************************************
 */
gint
Canvas::getBoardWidth() const
{
  const Geometry &geometry = getBoard()->getGeometry();
  return (geometry.getMaxNumber() * (CELL_SIZE + CELL_LINE_WIDTH))
         + geometry.getGridsPerRow() * (GRID_LINE_WIDTH - CELL_LINE_WIDTH)
         + GRID_LINE_WIDTH;
}


/*********************************************************************
 */
gint
Canvas::getBoardHeight() const
{
  const Geometry &geometry = getBoard()->getGeometry();
  return (geometry.getMaxNumber() * (CELL_SIZE + CELL_LINE_WIDTH))
         + geometry.getGridsPerColumn() * (GRID_LINE_WIDTH - CELL_LINE_WIDTH)
         + GRID_LINE_WIDTH;
}


/*********************************************************************
 */
void
Canvas::drawGridLines(GdkGC *gc, const Geometry &geometry)
{
  const Geometry::size_type segsize = geometry.getGridsPerRow()
                        + geometry.getGridsPerColumn()
                        + 2;
  auto_array_ptr<GdkSegment> segs( new GdkSegment[ segsize ] );
  GdkSegment *segment = segs.get();

  const gint boardWidth = getBoardWidth();
  gint height = TOP_BORDER_HEIGHT + (GRID_LINE_WIDTH / 2);
  Geometry::size_type i = 0;

  while (i <= geometry.getGridsPerColumn())
  {
    segment[i].x1 = LEFT_BORDER_WIDTH;
    segment[i].x2 = LEFT_BORDER_WIDTH + boardWidth;
    segment[i].y1 = segment[i].y2 = height;

    height += (geometry.getGridHeight() * (CELL_SIZE + CELL_LINE_WIDTH))
              + (GRID_LINE_WIDTH - CELL_LINE_WIDTH);
    ++i;
  } /* while */

  const gint boardHeight = getBoardHeight();
  gint width = LEFT_BORDER_WIDTH + (GRID_LINE_WIDTH / 2);

  while (i < segsize)
  {
    segment[i].y1 = TOP_BORDER_HEIGHT;
    segment[i].y2 = TOP_BORDER_HEIGHT + boardHeight;
    segment[i].x1 = segment[i].x2 = width;

    width += (geometry.getGridWidth() * (CELL_SIZE + CELL_LINE_WIDTH))
             + (GRID_LINE_WIDTH - CELL_LINE_WIDTH);
    ++i;
  } /* while */

  gdk_gc_set_line_attributes(gc, GRID_LINE_WIDTH,
                                 GDK_LINE_SOLID,
                                 GDK_CAP_BUTT,
                                 GDK_JOIN_MITER);
  gdk_draw_segments(m_backingStore, gc, segment, segsize);
}


/*********************************************************************
 */
void
Canvas::drawCellLines(GdkGC *gc, const Geometry &geometry)
{
  const Geometry::size_type horizontals
             = (geometry.getGridHeight() - 1) * geometry.getGridsPerColumn();
  const Geometry::size_type verticals
             = (geometry.getGridWidth() - 1) * geometry.getGridsPerRow();
  const Geometry::size_type segsize = horizontals + verticals;
  auto_array_ptr<GdkSegment> segs( new GdkSegment[ segsize ] );
  GdkSegment *segment = segs.get();

  const gint boardWidth = getBoardWidth();
  gint height = TOP_BORDER_HEIGHT + (CELL_LINE_WIDTH / 2);
  Geometry::size_type i = 0;
  unsigned n = 0;

  while (i < horizontals )
  {
    drawLabel(LEFT_BORDER_WIDTH/2, height + 4, n++);

    height += (CELL_SIZE + GRID_LINE_WIDTH);

    for (Geometry::size_type j = 1; j < geometry.getGridHeight(); ++j)
    {
      segment[i].x1 = LEFT_BORDER_WIDTH;
      segment[i].x2 = LEFT_BORDER_WIDTH + boardWidth;
      segment[i].y1 = segment[i].y2 = height;

      drawLabel(LEFT_BORDER_WIDTH/2, height + 4, n++);
      height += (CELL_SIZE + CELL_LINE_WIDTH);
      ++i;
    } /* for */
  } /* while */

  const gint boardHeight = getBoardHeight();
  gint width = LEFT_BORDER_WIDTH + (CELL_LINE_WIDTH / 2);
  n = 0;

  while (i < segsize)
  {
    drawLabel(width + CELL_SIZE/2, 2, n++);

    width += (CELL_SIZE + GRID_LINE_WIDTH);

    for (Geometry::size_type j = 1; j < geometry.getGridWidth(); ++j)
    {
      segment[i].y1 = TOP_BORDER_HEIGHT;
      segment[i].y2 = TOP_BORDER_HEIGHT + boardHeight;
      segment[i].x1 = segment[i].x2 = width;

      drawLabel(width + CELL_SIZE/2, 2, n++);
      width += (CELL_SIZE + CELL_LINE_WIDTH);
      ++i;
    } /* for */
  } /* while */

  gdk_gc_set_line_attributes(gc, CELL_LINE_WIDTH,
                                 GDK_LINE_SOLID,
                                 GDK_CAP_BUTT,
                                 GDK_JOIN_MITER);
  gdk_draw_segments(m_backingStore, gc, segment, segsize);
}


/*********************************************************************
 */
void
Canvas::drawGrid(GdkGC *gcGrid, GdkGC *gcCells)
{
  const Geometry &geometry = getBoard()->getGeometry();
  drawCellLines(gcCells, geometry);
  drawGridLines(gcGrid, geometry);
}


/*********************************************************************
 */
void
Canvas::drawLabel(int x, int y, unsigned n)
{
  PangoLayout *layout = getLayoutFor(n);
  if (layout != NULL)
  {
    pango_layout_set_alignment(layout, PANGO_ALIGN_CENTER);
    pango_layout_set_width(layout, LABEL_SIZE);

    PangoContext *context = pango_layout_get_context(layout);
    PangoFontDescription *fontDesc;

    fontDesc = pango_context_get_font_description(context);
    pango_font_description_set_weight(fontDesc, PANGO_WEIGHT_BOLD);
    pango_font_description_set_size(fontDesc, LABEL_FONT_SIZE*PANGO_SCALE);

    gdk_draw_layout( m_backingStore,
                     getGtkWidget()->style->fg_gc[GTK_STATE_NORMAL],
                     x,
                     y,
                     layout );

    g_object_unref(layout);
  }
}


/*********************************************************************
 */
void
Canvas::clear()
{
  if (m_backingStore != NULL)
    g_object_unref(m_backingStore);

  GtkWidget* widget = getGtkWidget();

  m_backingStore = gdk_pixmap_new(widget->window,
                                  widget->allocation.width,
                                  widget->allocation.height,
                                  -1);
  gdk_draw_rectangle(m_backingStore,
                     widget->style->white_gc,
                     TRUE,
                     0,
                     0,
                     widget->allocation.width,
                     widget->allocation.height);
  gdk_draw_rectangle(m_backingStore,
                     widget->style->black_gc,
                     FALSE,
                     0,
                     0,
                     widget->allocation.width - 1,
                     widget->allocation.height - 1);
  drawGrid( widget->style->fg_gc[GTK_STATE_NORMAL],
            widget->style->fg_gc[GTK_STATE_INSENSITIVE] );

  Geometry g = getBoard()->getGeometry();
}


/*********************************************************************
 */
void
Canvas::repaint()
{
  clear();
  show( *getDocument() );
  clean();
}


/*********************************************************************
 */
void
Canvas::resize()
{
  GtkWidget *w = getGtkWidget();
  const gint width = getCanvasWidth();
  const gint height = getCanvasHeight();

  if ((w->allocation.width != width) || (w->allocation.height != height))
  {
    gtk_widget_set_size_request(w, width, height);
    dirty();
  }
}


/*********************************************************************
 */
void
Canvas::refresh() const
{
  const GtkWidget *w = getGtkWidget();
  gtk_widget_queue_draw_area(const_cast<GtkWidget*>(w),
                             0,
                             0,
                             w->allocation.width,
                             w->allocation.height);
}


/*********************************************************************
 */
PangoLayout*
Canvas::getLayoutFor(unsigned n) const
{
  gchar *number = g_strdup_printf("%u", n);

  PangoLayout *layout = gtk_widget_create_pango_layout(getGtkWidget(), number);
  g_free(number);

  return layout;
}


/*********************************************************************
 */
void
Canvas::draw(const GdkRectangle &rect, unsigned n, const GdkColor *fg)
{
  if (n == 0)
  {
    gdk_draw_rectangle(m_backingStore,
                       getGtkWidget()->style->white_gc,
                       TRUE,
                       rect.x,
                       rect.y,
                       rect.width,
                       rect.height);
  }
  else
  {
    PangoLayout *layout = getLayoutFor(n);
    if (layout != NULL)
    {
      pango_layout_set_alignment(layout, PANGO_ALIGN_CENTER);
      pango_layout_set_width(layout, CELL_SIZE);

      PangoContext *context = pango_layout_get_context(layout);
      PangoFontDescription *fontDesc;

      fontDesc = pango_context_get_font_description(context);
      pango_font_description_set_weight(fontDesc, PANGO_WEIGHT_BOLD);
      pango_font_description_set_size(fontDesc, CELL_FONT_SIZE*PANGO_SCALE);

      gdk_draw_layout_with_colors(
                              m_backingStore,
                              getGtkWidget()->style->fg_gc[GTK_STATE_NORMAL],
                              rect.x + (CELL_SIZE / 2),
                              rect.y,
                              layout,
                              fg,
                              NULL);

      g_object_unref(layout);
    }
  }
}


/*********************************************************************
 */
void
Canvas::draw(const Square &square, const GdkColor *fg)
{
  GdkRectangle rect;
  draw(getCellRectangle(square, rect), square.getNumber(), fg);
}


/*********************************************************************
 */
void
Canvas::draw(const Board *board, const GdkColor *fg)
{
  if (board != NULL)
    const_cast<Board*>(board)->doAllSquares( Draw<Square>(this, fg) );
}


/*********************************************************************
 */
void
Canvas::drawOver(const Square &square, const GdkColor *fg)
{
  const unsigned number = square.getNumber();
  if (number != 0)
  {
    GdkRectangle rect;
    draw(getCellRectangle(square, rect), 0, NULL);
    draw(rect, number, fg);
  }
}


/*********************************************************************
 */
void
Canvas::drawOver(const Board *board, const GdkColor *fg)
{
  if (board != NULL)
    const_cast<Board*>(board)->doAllSquares( DrawOver<Square>(this, fg) );
}


/*********************************************************************
 * Write our changes to the backing store onto the window.
 */
void
Canvas::show(const Square &square, const GdkColor *fg)
{
  GdkRectangle rect;

  draw(getCellRectangle(square, rect), square.getNumber(), fg);

  gtk_widget_queue_draw_area(getGtkWidget(),
                             rect.x,
                             rect.y,
                             rect.width,
                             rect.height);
}


/*********************************************************************
 */
void
Canvas::show(const Board *board, const GdkColor *fg)
{
  if (board != NULL)
  {
    const_cast<Board*>(board)->doAllSquares( Draw<Square>(this, fg) );
    refresh();
  }
}


/*********************************************************************
 */
void
Canvas::show(const Document &document)
{
  draw( document.getOriginalBoard() );
  document.do_for_each_hint( DrawRef(this, &RED) );
  document.do_for_each_move( DrawRef(this, &BLUE) );
  refresh();
}


/*********************************************************************
 */
gint
Canvas::getCellDistance(gint index, gint gridSize)
{
  return LEFT_BORDER_WIDTH +
         GRID_LINE_WIDTH +
         (index * (CELL_SIZE + CELL_LINE_WIDTH)) +
         ((index / gridSize) * (GRID_LINE_WIDTH - CELL_LINE_WIDTH));
}


/*********************************************************************
 */
gint
Canvas::getCellYDistance(const Square &square)
{
  return getCellDistance(square.getRowIndex(),
                         square.getGeometry().getGridHeight());
}


/*********************************************************************
 */
gint
Canvas::getCellXDistance(const Square &square)
{
  return getCellDistance(square.getColumnIndex(),
                         square.getGeometry().getGridWidth());
}


/*********************************************************************
 */
GdkRectangle&
Canvas::getCellRectangle(const Square &square, GdkRectangle &rect)
{
  rect.x = getCellXDistance(square);
  rect.y = getCellYDistance(square);

  rect.width = rect.height = CELL_SIZE;
  
  return rect;
}


/*********************************************************************
 */
gint
Canvas::getCellRowIndex(gint y, const Geometry &geometry)
{
  const gint gridHeight
               = (CELL_SIZE + CELL_LINE_WIDTH) * geometry.getGridHeight()
                 + (GRID_LINE_WIDTH - CELL_LINE_WIDTH);
  return (y - (y / gridHeight) * (GRID_LINE_WIDTH - CELL_LINE_WIDTH))
            / (CELL_SIZE + CELL_LINE_WIDTH);
}


/*********************************************************************
 */
gint
Canvas::getCellColumnIndex(gint x, const Geometry &geometry)
{
  const gint gridWidth
               = (CELL_SIZE + CELL_LINE_WIDTH) * geometry.getGridWidth()
                 + (GRID_LINE_WIDTH - CELL_LINE_WIDTH);
  return (x - (x / gridWidth) * (GRID_LINE_WIDTH - CELL_LINE_WIDTH))
            / (CELL_SIZE + CELL_LINE_WIDTH);
}


/*********************************************************************
 */
void
Canvas::restoreDefaults()
{
  resize();
}

}

