#include <ctime>
#include <sstream>
#include <iomanip>
using namespace std;

#include "document.h"
#include "geometry.h"
#include "solution.h"
#include "actions.h"
#include "generator.h"

extern "C"
{
#  include "support.h"
}

namespace sudoku
{

/*********************************************************************
 */
class Write : public Document::SquareRefAction
{
private:
  ostream& m_ostr;
  const char *m_tag;

public:
  Write(ostream &ostr, const char *tag) : m_ostr(ostr), m_tag(tag) {}

  void operator()(Square *s) const {
      m_ostr << m_tag << ": "
             << s->getSquareIndex()
             << ","
             << s->getNumber()
             << endl;
  }
};


/*********************************************************************
 */
void
Document::do_for_each_move(const SquareRefAction &action) const
{
  do_for_each(m_moves.getMoves().begin(),
              m_moves.getMoves().end(),
              action);
}


/*********************************************************************
 */
void
Document::do_for_each_hint(const SquareRefAction &action) const
{
  do_for_each(m_givenHints.begin(), m_givenHints.end(), action);
}


#if 0
void
Document::do_for_each(const RefCollection &items,
                      const RefAction &action)
{
  const RefCollection::const_iterator end = items.end();
  RefCollection::const_iterator square = items.begin();

  while (square != end)
  {
    action(*square);
    ++square;
  } /* while */
}
#endif


/*********************************************************************
 */
void
Document::throw_illegal_option(const char *tag, const Square &square)
{
  string msg(tag);
  throw runtime_error( msg.append(": ").append(square.getName()) );
}


/*********************************************************************
 */
Document::Document()
  : m_original( NULL )
  , m_board( new Board( Geometry() ) )
  , m_hasLogical(false)
  , m_hint( m_allHints.end() )
  , m_hintsAreStale(true)
{}


/*********************************************************************
 */
Document::~Document()
{
  delete m_original;
  delete m_board;
}


/*********************************************************************
 */
Board&
Document::recreate(const Geometry &geometry)
{
  m_moves.clear();
  m_givenHints.clear();

  m_hasLogical = false;

  delete m_original;
  m_original = NULL;

  delete m_board;
  m_board = new Board(geometry);
  return *m_board;
}


/*********************************************************************
 */
void
Document::clear()
{
  staleHints();
  m_board->clear();
}


/*********************************************************************
 */
void
Document::finalise()
{
  delete m_original;
  m_original = new Board( *m_board );

  solve();
  m_original->setRating( m_board->getRating() );
}


/*********************************************************************
 */
void
Document::generate(int index)
{
  const Geometry geometry( getGeometry() );

  Board& board = recreate( geometry );
  do
  {
    clear();
    Generator().generate(board);
    finalise();
  }
  while (Board::getRatingIndex( board.getRating() ) < index);
}


/*********************************************************************
 */
void
Document::loadPlayState(istream &istr)
{
  const Board::SquareCollection::size_type
                                  maxSquares = getBoard()->getSquares().size();

  StepCollection  hints;
  StepCollection  moves;

  StringPair option;
  string line;

  while ( !getLine(istr, line).empty() )
  {
    TRACE( cout << "Line: [" << line << "]" << endl; )

    if ( getOption(line, option) )
    {
      bool failed = true;

      StepCollection *collect = NULL;

      if (option.name.compare("SQUARE") == 0)
        collect = &moves;
      else if (option.name.compare("HINT") == 0)
        collect = &hints;

      if (collect != NULL)
      {
        istringstream is(option.value);
        unsigned squareIndex, number;
        char c;

        is >> squareIndex >> ws >> c >> ws >> number;
        if (is && (squareIndex < maxSquares) && (number != 0))
        {
          collect->push_back( Step(squareIndex, number) );
          failed = false;
        }
      }

      if (failed)
      {
        string msg( _("Invalid option [") );
        throw runtime_error(msg.append(line).append("]"));
      }
    }
  } /* while */

  /*
   * Now load the hints into the board, followed by the moves.
   * We need to load things in this order, regardless of the
   * order they appear in the input file, so that the player
   * isn't warned prematurely of any errors s/he might have
   * made.
   */
  StepCollection::const_iterator end = hints.end();
  StepCollection::const_iterator step = hints.begin();

  while (step != end)
  {
    Square &square = getBoard()->find(step->squareIndex);
    if (square.getNumber() != 0)
      throw_illegal_option("Hint", square);

    addHint(square, step->number);
    ++step;
  } /* while */

  end = moves.end();
  step = moves.begin();

  while (step != end)
  {
    Square &square = getBoard()->find(step->squareIndex);
    if (square.getNumber() != 0)
      throw_illegal_option("Move", square);

    addMove(square, step->number);
    ++step;
  } /* while */

  staleHints();
}


/*********************************************************************
 */
void
Document::load(istream &istr)
{
  Geometry geometry;

  istr >> geometry;
  istr >> recreate(geometry);
  finalise();

  /*
   * Now that we've finalised the original puzzle,
   * let's reload any moves which may have been
   * made already ...
   */
  loadPlayState(istr);
}


/*********************************************************************
 */
void
Document::savePuzzle(ostream &ostr) const
{
  if (m_original != NULL)
  {
    ostr << Board::Exporter(*m_original);
  }
}


/*********************************************************************
 */
void
Document::savePlay(ostream &ostr) const
{
  savePuzzle(ostr);
  ostr << endl;

  do_for_each_hint( Write(ostr, "Hint") );
  do_for_each_move( Write(ostr, "Square") );
}

 
/*********************************************************************
 */
bool
Document::solve(Board &board,
                BoardCollection &solutions,
                StepCollection *solutionLog)
{
  Puzzle puzzle(board);
  bool logical;

  TRACE( cout << "DOCUMENT:: SOLVING" << endl; )

  solutions.clear();
  if ( (logical = puzzle.solve()) )
  {
    solutions.push_back(board);
  }
  else
  {
    TRACE( cout << "DOCUMENT::  FINDING ALL SOLUTIONS" << endl; )
    puzzle.findAllSolutions( solutions, UPPER_SOLUTION_LIMIT );
  }

  if (solutionLog != NULL)
    puzzle.getSolution(*solutionLog);

  TRACE( cout << "DOCUMENT:: Logical solution? "
              << (logical ? "Y" : "N")
              << endl; )
  return logical;
}


/*********************************************************************
 */
bool
Document::solve()
{
  Board board( *getBoard() );
  m_hasLogical = solve(board, m_solutions, &m_solutionLog);
  getBoard()->setRating( board.getRating() );
  return m_hasLogical;
}


/*********************************************************************
 */
bool
Document::validate(const Board *b, BoardCollection &solutions)
{
  TRACE( b->dump(cout, "Validating-A"); )
  Board board(*b);
  TRACE( board.dump(cout, "Validating-B"); )
  return solve(board, solutions);
}


/*********************************************************************
 */
const char*
Document::getRating(const BoardCollection &solutions)
{
  BoardCollection::const_iterator first = solutions.begin();
  const Board *board
               = ((first != solutions.end()) ? first.operator->() : NULL);
  return (board != NULL ? board->getRating( board->getRating() ) : "");
}


/*********************************************************************
 */
string&
Document::generateFileName(string &name) const
{
  ostringstream ostr;
  struct tm date;
  time_t now;

  now = time(NULL);
  localtime_r(&now, &date);

  ostr << "Generated-" << (1900 + date.tm_year)
                       << setfill('0')
                       << setw(2) << (1 + date.tm_mon)
                       << setw(2) << date.tm_mday
                       << setw(2) << date.tm_hour
                       << setw(2) << date.tm_min
                       << setw(2) << date.tm_sec
                       << "-"
                       << getRating();

  return (name = ostr.str());
}


/*********************************************************************
 */
const string&
Document::getStatus(string &str, bool hasLogical, size_type count)
{
  if (hasLogical)
  {
    str = _("Unique logical solution");
  }
  else
  {
    ostringstream ostr;

    switch (count)
    {
    case 1:
      ostr << _("Unique solution (with some guesswork)");
      break;

    case UPPER_SOLUTION_LIMIT:
      ostr << _("At least ") << count << _(" solutions");
      break;

    default:
      ostr << count << _(" solutions");
      break;
    } /* switch */

    str = ostr.str();
  }

  return str;
}


/*********************************************************************
 */
bool
Document::isComplete(const Board *board)
{
  bool complete = true;

  const SquareCollection::const_iterator end = board->getSquares().end();
  SquareCollection::const_iterator square = board->getSquares().begin();

  while (square != end)
  {
    if (square->getNumber() == 0)
    {
      complete = false;
      break;
    }

    ++square;
  } /* while */

  return complete;
}


/*********************************************************************
 */
void
Document::addHint(Square &square, unsigned number)
{
  square = Square::FIXED(number);
  m_givenHints.push_back( &square );
}


/*********************************************************************
 */
void
Document::addMove(Square &square, unsigned number)
{
  /*
   * We want to allow our player to make SOME mistakes!
   * This function will not alert us if we accidentally
   * make it impossible to place any number in a particular
   * square.
   */
  square.setNumber_player( number );

  m_moves.add(&square);
  staleHints();
}


/*********************************************************************
 */
Square*
Document::removeLatestMove()
{
  Square *square = m_moves.getLatest();

  if (square != NULL)
  {
    square->revertNumber();
    m_moves.removeLatest();
    staleHints();
  }

  return square;
}

/*********************************************************************
 */
const Hint*
Document::getHint()
{
  const Hint *hint = NULL;

  if ( hintsAreStale() )
    generateHints();

  if (m_hint != m_allHints.end())
  {
    hint = &(*m_hint);
    ++m_hint;
  }

  return hint;
}


/*********************************************************************
 */
void
Document::generateHints()
{
  try
  {
    Board board( *m_board );
    BoardCollection solutions;

    solve(board, solutions, &m_allHints);

    if (solutions.empty())
    {
      m_allHints.clear();
      m_allHints.push_back(
              Comment(_("Cannot complete the puzzle from this position.")) );
    }
  }
  catch(const violation &e)
  {
    m_allHints.clear();

    string str(e.what());
    str.append( _(": NO SOLUTION") );
    m_allHints.push_back( Comment(str) );
  }

  m_hint = m_allHints.begin();
  freshHints();
}

}

