#ifndef DOCUMENT_H
#define DOCUMENT_H

#include <string>
#include <istream>

#include "puzzle.h"
#include "board.h"
#include "actions.h"

namespace sudoku
{
  class Document
  {
  public:
    typedef Board::SquareCollection  SquareCollection;
    typedef BoardCollection::size_type  size_type;

    typedef Puzzle::SquareCollection  RefCollection;
    typedef RefAction<Square>         SquareRefAction;

    class Moves
    {
    private:
      RefCollection  m_moves;

    public:
      void add(Square *s) { m_moves.push_back(s); }
      void removeLatest() { m_moves.pop_back(); }
      Square* getLatest() const {
          return (m_moves.empty() ? NULL : m_moves.back());
      }
      void clear() { m_moves.clear(); }
      size_type count() const { return m_moves.size(); }

      const RefCollection& getMoves() const { return m_moves; }
    };

  private:
    BoardCollection m_solutions;
    Board*          m_original;
    Board*          m_board;

    bool            m_hasLogical;
    StepCollection  m_solutionLog;

    Moves           m_moves;
    RefCollection   m_givenHints;

    HintCollection  m_allHints;
    HintIterator    m_hint;

    bool            m_hintsAreStale;

    bool hintsAreStale() const { return m_hintsAreStale; }
    void staleHints() { m_hintsAreStale = true; }
    void freshHints() { m_hintsAreStale = false; }

    static bool solve(Board&, BoardCollection&, StepCollection* =NULL);
    bool solve();

    void loadPlayState(std::istream &istr);
    void generateHints();

    static void throw_illegal_option(const char*, const Square&);

  public:
    static const size_type UPPER_SOLUTION_LIMIT = 5;

    Document();
    ~Document();

    Board* getBoard() const { return m_board; }
    Board* getOriginalBoard() const { return m_original; }
    const Geometry& getGeometry() const { return m_board->getGeometry(); }

    const BoardCollection& getSolutions() const { return m_solutions; }
    size_type getSolutionCount() const { return m_solutions.size(); }
    bool hasLogicalSolution() const { return m_hasLogical; }

    const StepCollection& getSolutionLog() const { return m_solutionLog; }

    const char* getRating() const { return getRating(m_solutions); }
    const std::string& getStatus(std::string &s) const {
        return getStatus(s, hasLogicalSolution(), getSolutionCount());
    }

    bool isComplete() const {
        return isComplete( getBoard() );
    }

    Board& recreate(const Geometry&);
    void load(std::istream&);
    void savePlay(std::ostream&) const;
    void savePuzzle(std::ostream&) const;
    void clear();
    void finalise();
    void generate(int rating=0);

    void addMove(Square&, unsigned);
    Square* removeLatestMove();
    size_type countMoves() const { return m_moves.count(); }

    void do_for_each_move(const SquareRefAction &action) const;
    void do_for_each_hint(const SquareRefAction &action) const;

    const Hint *getHint();
    bool hasHints() const { return (m_hint != m_allHints.end()); }
    void addHint(Square&, unsigned);

    std::string& generateFileName(std::string&) const;
    static const char* getRating(const BoardCollection&);
    static bool validate(const Board*, BoardCollection&);
    static const std::string& getStatus(std::string&, bool, size_type);
    static bool isComplete(const Board*);
  };
}

#endif

