#include <iterator>
#include <iostream>
#include <sstream>

#include "puzzle.h"
#include "board.h"
#include "freedom.h"
#include "square.h"
#include "solution.h"
#include "actions.h"
#include "selector.h"
#include "strategy.h"
using namespace std;


namespace sudoku
{

/*********************************************************************
 */
Puzzle::Puzzle(Board &board)
  : m_board(board)
  , m_forcedSquareScore(DEFAULT_FORCED_SQUARE_SCORE)
  , m_completedSequenceScore(DEFAULT_COMPLETED_SEQUENCE_SCORE)
  , m_crossSequenceConstraintScore(DEFAULT_CROSS_SEQUENCE_CONSTRAINT_SCORE)
  , m_numberSubsetScore(DEFAULT_NUMBER_SUBSET_SCORE)
  , m_guessworkScore(DEFAULT_GUESSWORK_SCORE)
  , m_guessSelector(NULL)
{
  Board::SquareCollection::iterator square = board.getSquares().begin();
  const Board::SquareCollection::iterator endSquare = board.getSquares().end();

  /*
   * We want to keep a list of all squares that haven't
   * yet been assigned a number ...
   */
  while (square != endSquare)
  {
    if (square->getNumber() == 0)
    {
      m_squares.push_back( square.operator->() );
    }

    ++square;
  } /* while */

  Board::SequenceCollection::iterator sequence = board.getSequences().begin();
  const Board::SequenceCollection::iterator endSequence = board.getSequences().end();

  /*
   * We also want to keep a list of every sequence that
   * has at least one hole in it ...
   */
  while (sequence != endSequence)
  {
    if (sequence->getMissing().getCount() != 0)
    {
      m_sequences.push_back( sequence.operator->() );
    }

    ++sequence;
  } /* while */

  /*
   * These are the strategies that we will use
   * on our sequences ...
   */
  m_strategies.push_back( Strategy::FindCompleteSequences );

  /*
   * And these are the strategies that we are keeping in reserve ...
   */
  m_newStrategies.push_back( Strategy::FindCrossSequenceConstraints );
  m_newStrategies.push_back( Strategy::FindNumberSubsets );

  /*
   * And this is the top of our "undo" stack ...
   */
  getStack().push_back( State() );
}


/*********************************************************************
 */
Puzzle::~Puzzle()
{
  delete m_guessSelector;
}


/*********************************************************************
 */
#ifdef DEBUG
void
Puzzle::dumpBoard(ostream &ostr, const string &tag) const
{
  getBoard().dump(ostr, tag);
}
#else
inline void
Puzzle::dumpBoard(ostream&, const string&) const {}
#endif


/*********************************************************************
 */
void
Puzzle::completeSequence(SequenceCollection::iterator &iter)
{
  getStack().back().fullSequences.push_back( *iter );
  getSequences().erase(iter);
}


/*********************************************************************
 */
void
Puzzle::completeSquare(SquareCollection::iterator &iter)
{
  getStack().back().fullSquares.push_back( *iter );
  getSquares().erase(iter);
}


/*********************************************************************
 */
void
Puzzle::push()
{
  getStack().push_back( State() );
  getBoard().push();

  dumpBoard(cout, "SAVED");
}


/*********************************************************************
 */
bool
Puzzle::pop()
{
  bool ok = false;

  if ( !getStack().empty() )
  {
    /*
     * The sequences and squares that we completed since the last
     * "push" must now be restored ...
     */
    State& state = getStack().back();

    copy(state.fullSequences.begin(), state.fullSequences.end(),
         front_insert_iterator<SequenceCollection>(getSequences()));
    copy(state.fullSquares.begin(), state.fullSquares.end(),
         front_insert_iterator<SquareCollection>(getSquares()));
    getStack().pop_back();

    /*
     * Now restore the board itself ...
     */
    ok = getBoard().pop();

    dumpBoard(cout, "RESTORED");
  }

  return ok;
}


/*********************************************************************
 */
void
Puzzle::addForcedSquare(const Square *square, unsigned n)
{
  m_solution.push_back( ForcedSquare(square, n) );
  m_rating.addScore( getForcedSquareScore() );
}


/*********************************************************************
 */
int
Puzzle::setForcedSquares()
{
  TRACE( cout << "FORCING SQUARES" << endl; )

  int changes = 0;

  SquareCollection::iterator sqiter = getSquares().begin();
  const SquareCollection::iterator end = getSquares().end();

  while (sqiter != end)
  {
    bool killSquare = false;

    Square *square = *sqiter;

    if (square->getNumber() != 0)
    {
      killSquare = true;
    }
    else if (square->getPossibilities() == 1)
    {
      const bitmask_t forbidden = square->getForbidden();
      bitmask_t mask = 1;

      for (unsigned n = 1; n <= getBoard().getMaxNumber(); ++n)
      {
        if ((forbidden & (mask <<= 1)) == 0)
        {
          addForcedSquare(square, n);

          square->setNumber(n);
          ++changes;

          killSquare = true;
          break;
        }
      } /* for */
    }

    if (killSquare)
    {
      SquareCollection::iterator deadIter = sqiter--;
      completeSquare(deadIter);
    }

    ++sqiter;
  } /* while */

  return changes;
}


/*********************************************************************
 */
bool
Puzzle::getDegreesOfFreedom(Freedom &freedom) const
{
  const Sequence *sequence = freedom.getSequence();

  const SquareSequence::const_iterator endSquare = sequence->getSquares().end();
  SquareSequence::const_iterator sqiter = sequence->getSquares().begin();
  bool hasFreedom = false;
  unsigned sq = 0;

  TRACE( cout << "Getting freedoms for sequence "
              << sequence->getName()
              << endl; )

  while (sqiter != endSquare)
  {
    const Square *square = *sqiter;

    const bitmask_t forbiddenMask = square->getForbidden();
    bitmask_t numberMask = 1;

    for (Freedom::size_type n = 1; n < freedom.size(); ++n)
    {
      numberMask <<= 1;
      if ( ((forbiddenMask & numberMask) == 0) &&
           (square->getNumber() == 0) )
      {
        freedom[n].setBit(sq);
        hasFreedom = true;
      }
    } /* for */

    ++sq;
    ++sqiter;
  } /* while */

  return hasFreedom;
}


/*********************************************************************
 */
void
Puzzle::addCompletedSequence(const Square *square,
                             const Sequence *sequence,
                             unsigned n)
{
  m_solution.push_back( CompletedSequence(square, n, sequence) );
  m_rating.addScore( getCompletedSequenceScore() );
}


/*********************************************************************
 */
void
Puzzle::addNumberSubset(const Sequence *sequence,
                        const Subset &subset)
{
  m_solution.push_back( NumberSubset(sequence, subset) );
  m_rating.addScore( getNumberSubsetScore() );
}


/*********************************************************************
 */
void
Puzzle::addCrossSequenceConstraint(const Square *square,
                                   const Sequence *sequence,
                                   const Sequence *cross,
                                   unsigned n)
{
  m_solution.push_back(
               CrossSequenceConstraint(square, n, sequence, cross) );
  m_rating.addScore( getCrossSequenceConstraintScore() );
}


/*********************************************************************
 */
void
Puzzle::addGuesswork(const Square *square, unsigned n)
{
  m_solution.push_back( Guesswork(square, n) );
  m_rating.addScore( getGuessworkScore() );
}


/*********************************************************************
 */
int
Puzzle::processSequenceStrategy(const Strategy &strategy)
{
  int changes = 0;

  SequenceCollection::iterator iter = getSequences().begin();
  const SequenceCollection::iterator end = getSequences().end();

  while (iter != end)
  {
    Sequence *sequence = *iter;

    if (sequence->getMissing().getCount() != 0)
    {
      Freedom freedom(sequence);
      if (getDegreesOfFreedom(freedom))
      {
        changes += strategy(this, freedom);
      }
    }

    if (sequence->getMissing().getCount() == 0)
    {
      SequenceCollection::iterator deadIter = iter--;
      completeSequence(deadIter);
    }

    ++iter;
  } /* while */

  return changes;
}


/*********************************************************************
 */
int
Puzzle::includeNewStrategy()
{
  int changes = 0;

  if ( !getNewStrategies().empty() )
  {
    const StrategyCollection::iterator end = getNewStrategies().end();
    StrategyCollection::iterator strategy = getNewStrategies().begin();

    while (strategy != end)
    {
      changes = processSequenceStrategy(*strategy);
      if (changes != 0)
      {
        TRACE( cout << "Added new sequence strategy" << endl; )

        getStrategies().push_back(*strategy);
        getNewStrategies().erase(strategy);
        break;
      }

      ++strategy;
    } /* while */
  }

  return changes;
}


/*********************************************************************
 */
int
Puzzle::processSequences()
{
  int changes = processSequenceStrategy( Strategy::ApplyAll );

  if (changes == 0)
    changes = includeNewStrategy();

  return changes;
}


/*********************************************************************
 */
bool
Puzzle::solveByDeduction()
{
  /*
   * We have two strategies, and one is much cheaper
   * than the other. So we use the cheaper one until
   * it can no longer help us, and then we try using
   * the more expensive one. If that helps us to
   * progress then we switch back to using the cheaper
   * one again. With any luck, we'll have completed
   * the puzzle by the time both strategies fail us.
   */
  do
  {
    while (setForcedSquares() != 0) {}
  }
  while ( !getSquares().empty() &&
          (processSequences() != 0) &&
          !getSequences().empty() );

  return (getSquares().empty() || getSequences().empty());
}


/*********************************************************************
 */
bool
Puzzle::guess(Square *square, unsigned n)
{
  bool ok = false;

  TRACE( cout << "Guessing " << square->getName() << " is " << n << endl; )
  try
  {
    if ( square->forbid_nothrow(n) )
    {
      if (square->getPossibilities() != 0)
        push();

      square->setNumber_checkSeqs(n);
      addGuesswork(square, n);
      ok = true;
    }
  }
  catch(const violation&)
  {
    pop();
  }

  return ok;
}


/*********************************************************************
 */
bool
Puzzle::solveByBruteForce(const Selector &selectNumber)
{
  TRACE( cout << "GOING FOR BRUTE FORCE" << endl; )

  getSquares().sort( FewestPossibilitiesFirst() );

  const SquareCollection::iterator end = getSquares().end();
  SquareCollection::iterator sq = getSquares().begin();

  while (sq != end)
  {
    Square *square = *sq;

    if ((square->getNumber() != 0) || (square->getPossibilities() == 0))
    {
      SquareCollection::iterator deadIter = sq--;
      completeSquare(deadIter);
    }
    else
    {
      unsigned n = 0;

      /*
       * The number selector is an iterator of sorts:
       * We start with zero, and carry on until it
       * returns zero again.
       */
      while ( (n = selectNumber(square, n)) != 0 )
      {
        /* Guessing a square means that we haven't
         * finished "puzzling" yet ...
         */
        if ( guess(square, n) )
          return false;
      } /* while */
    }

    ++sq;
  } /* while */

  /*
   * We must have solved the puzzle, so we're done.
   */
  return true;
}


/*********************************************************************
 */
bool
Puzzle::solve()
{
  if ( completed() )
    return false;

  TRACE( cout << "BEGIN SOLVING" << endl; )

  bool isComplete = false;

  while ( !isComplete )
  {
    try
    {
      isComplete = solveByDeduction();
    }
    catch(const violation &e)
    {
      TRACE( cout << "VIOLATION: " << e.what() << endl; )

      ostringstream ostr;
      ostr << _("VIOLATION at ")
           << e.getSquare()->getName()
           << _(": Reversing most recent proposal");

      m_solution.push_back( Comment(ostr.str()) );
      isComplete = false;

      // Back-track to next position ...
      if ( !pop() )
        break;
    }

    if ( isComplete )
    {
      TRACE( cout << "COMPLETE" << endl; )
      break;
    }

    if ( m_guessSelector == NULL )
      break;

    isComplete = solveByBruteForce( *m_guessSelector );
  } /* while */

  if (isComplete)
  {
    m_board.setRating( getRating() );
  }

  TRACE( cout << "SOLVE EXITING: " << (isComplete ? "T" : "F") << endl; )
  return isComplete;
}


/*********************************************************************
 */
BoardCollection::size_type
Puzzle::findAllSolutions(BoardCollection &solutions,
                         BoardCollection::size_type maximum)
{
  const Selector *oldSelector = guessBy( new FirstAvailableNumber() );
  try
  {
    /*
     * "Zero maxmimum" is used to mean "Find every solution".
     */
    if (maximum == 0)
      --maximum;

    while ( (solutions.size() != maximum) && solve() )
    {
      solutions.push_back( getBoard() );

      ostringstream ostr;
      ostr << _("FOUND SOLUTION ") << solutions.size();
      m_solution.push_back( Comment(ostr.str()) );

      if ( !pop() )
        break;
    } /* while */

    delete guessBy(oldSelector);
  }
  catch(...)
  {
    delete guessBy(oldSelector);
  }

  return solutions.size();
}


/*********************************************************************
 */
void
Puzzle::getSolution(StepCollection &steps)
{
  steps = m_solution;
  m_solution.clear();
}


/*********************************************************************
 */
ostream&
Puzzle::showSolution(ostream &ostr) const
{
  ostream_iterator<StepCollection::value_type> citer(ostr, "\n");
  copy(getSolution().begin(), getSolution().end(), citer);
  return ostr;
}


/*********************************************************************
 */
int
Puzzle::getRating() const
{
  /*
   * Every time we placed a number, we increased our score.
   * Pro-rate this total score against the total number of
   * squares on the board.
   */
  return (m_rating.getScore() * m_rating.getMoves()) /
                                        getBoard().getSquares().size();
}


/*********************************************************************
 */
void
Puzzle::Rating::addScore(int score)
{
  if (score != 0)
  {
    m_score += score;
    ++m_moves;
  }
}

}

