#ifndef PUZZLE_H
#define PUZZLE_H

#include <ostream>
#include <list>

extern "C"
{
#  include "support.h"
}

namespace sudoku
{
  class Board;
  class Square;
  class Sequence;
  class Freedom;
  class Subset;
  class Selector;
  class Strategy;

  struct Step;
  typedef std::list<Step>  StepCollection;

  typedef Step  Hint;
  typedef StepCollection  HintCollection;
  typedef HintCollection::const_iterator  HintIterator;

  typedef std::list<Strategy>  StrategyCollection;

  typedef std::list<Board>  BoardCollection;

  class Puzzle
  {
  public:
    typedef std::list<Sequence*>  SequenceCollection;
    typedef std::list<Square*>    SquareCollection;

    static const int DEFAULT_FORCED_SQUARE_SCORE = 1;
    static const int DEFAULT_COMPLETED_SEQUENCE_SCORE = 2;
    static const int DEFAULT_CROSS_SEQUENCE_CONSTRAINT_SCORE = 3;
    static const int DEFAULT_NUMBER_SUBSET_SCORE = 7;
    static const int DEFAULT_GUESSWORK_SCORE = 10;

  private:
    struct State
    {
      SquareCollection    fullSquares;
      SequenceCollection  fullSequences;
    };

    class Rating
    {
    private:
      int m_score;
      int m_moves;

    public:
      Rating() : m_score(0), m_moves(0) {}

      void addScore(int);

      int getMoves() const { return m_moves; }
      int getScore() const { return m_score; }
    };

    typedef std::list<State>  StateStack;

    Board&              m_board;

    SquareCollection    m_squares;
    SequenceCollection  m_sequences;
    StateStack          m_stack;

    StrategyCollection  m_strategies;
    StrategyCollection  m_newStrategies;

    StepCollection      m_solution;
    Rating              m_rating;

    int                 m_forcedSquareScore;
    int                 m_completedSequenceScore;
    int                 m_crossSequenceConstraintScore;
    int                 m_numberSubsetScore;
    int                 m_guessworkScore;

    const Selector*     m_guessSelector;

    /* The default copy-constructor would not
     * handle all the object references correctly.
     */
    Puzzle(const Puzzle&);

    void completeSquare(SquareCollection::iterator&);
    void completeSequence(SequenceCollection::iterator&);

    const Board& getBoard() const { return m_board; }
    Board& getBoard() { return m_board; }

    const StateStack& getStack() const { return m_stack; }
    StateStack& getStack() { return m_stack; }

    bool guess(Square*, unsigned);

    void push();
    bool pop();

    bool getDegreesOfFreedom(Freedom&) const;
    int processSequenceStrategy(const Strategy&);
    int includeNewStrategy();
    int processSequences();
    int setForcedSquares();

  public:
    explicit Puzzle(Board &board);
    ~Puzzle();

    const Selector* getGuessSelector() const { return m_guessSelector; }
    const Selector* guessBy(const Selector *s) {
        const Selector *old = m_guessSelector;
        m_guessSelector = s;
        return old;
    }

    const SquareCollection& getSquares() const { return m_squares; }
    SquareCollection& getSquares() { return m_squares; }

    const SequenceCollection& getSequences() const { return m_sequences; }
    SequenceCollection& getSequences() { return m_sequences; }

    const StepCollection& getSolution() const { return m_solution; }
    StepCollection& getSolution() { return m_solution; }
    void getSolution(StepCollection&);

    const StrategyCollection& getStrategies() const { return m_strategies; }
    StrategyCollection& getStrategies() { return m_strategies; }

    const StrategyCollection& getNewStrategies() const {
      return m_newStrategies;
    }
    StrategyCollection& getNewStrategies() { return m_newStrategies; }

    bool solve();
    bool solveByDeduction();
    bool solveByBruteForce(const Selector&);
    bool completed() const { return getStack().empty(); }

    BoardCollection::size_type findAllSolutions(BoardCollection&,
                                                BoardCollection::size_type = 0);

    void addForcedSquare(const Square*, unsigned);
    void addCompletedSequence(const Square*, const Sequence*, unsigned);
    void addCrossSequenceConstraint(const Square*, const Sequence*,
                                    const Sequence*, unsigned);
    void addNumberSubset(const Sequence*, const Subset&);
    void addGuesswork(const Square*, unsigned);

    int getForcedSquareScore() const { return m_forcedSquareScore; }
    int getCompletedSequenceScore() const { return m_completedSequenceScore; }
    int getCrossSequenceConstraintScore() const {
          return m_crossSequenceConstraintScore;
    }
    int getNumberSubsetScore() const { return m_numberSubsetScore; }
    int getGuessworkScore() const { return m_guessworkScore; }

    void setForcedSquareScore(int s) { m_forcedSquareScore = s; }
    void setCompletedSequenceScore(int s) { m_completedSequenceScore = s; }
    void setCrossSequenceConstraintScore(int s) {
           m_crossSequenceConstraintScore = s; 
    }
    void setNumberSubsetScore(int s) { m_numberSubsetScore = s; }
    void setGuessworkScore(int s) { m_guessworkScore = s; }

    std::ostream& showSolution(std::ostream &ostr) const;

    int getScore() const { return m_rating.getScore(); }
    int getRating() const;

    void dumpBoard(std::ostream &, const std::string &) const;
  };
}

namespace std
{
  inline ostream&
  operator<<(ostream &ostr, const sudoku::Puzzle &puzzle) {
      return puzzle.showSolution(ostr);
  }
}

#endif

