#ifndef SQUARE_H
#define SQUARE_H

#include <string>
#include <list>

#include "constants.h"
#include "geometry.h"

namespace sudoku
{
  class Sequence;
  class Bitmask;

  class Square
  {
  public:
    class FIXED
    {
    private:
      unsigned m_number;

    public:
      explicit FIXED(unsigned n) : m_number(n) {}
      unsigned getNumber() const { return m_number; }
    };

  private:
    typedef void(Square::*SetterFn)(unsigned);

    struct State
    {
      bitmask_t forbidden;
      unsigned  possibilities;
      unsigned  number;

      State(const Geometry &g)
        : forbidden(0)
        , possibilities(g.getMaxNumber())
        , number(0)
      {}

      void clear(const Geometry&);
      void set(const State&);
    };

    typedef std::list<State>  StateStack;

    unsigned m_index;
    std::string m_name;

    const Geometry &m_geometry;
    mutable unsigned m_tag;

    State m_state;
    StateStack m_stack;

    Sequence *m_row;
    Sequence *m_column;
    Sequence *m_grid;

    const State& getState() const { return m_state; }
    State& getState() { return m_state; }
    StateStack& getStack() { return m_stack; }

    void setRow(Sequence *sequence) { m_row = sequence; }
    void setColumn(Sequence *sequence) { m_column = sequence; }
    void setGrid(Sequence *sequence) { m_grid = sequence; }

    void forbidAll();
    void copyState(const Square &s);

    bool setNumber_validator(SetterFn, unsigned);
    void setNumber_nothrow(unsigned newNumber);

    static std::string makeName(const Geometry&, unsigned sq);

  public:
    const std::string& getName() const { return m_name; }
    unsigned getSquareIndex() const { return m_index; }

    unsigned getNumber() const { return m_state.number; }
    void setNumber_checkSeqs(unsigned newNumber);
    bool setNumber(unsigned newNumber) {
        return setNumber_validator(&Square::setNumber_checkSeqs, newNumber);
    }
    bool setNumber_player(unsigned newNumber) {
        return setNumber_validator(&Square::setNumber_nothrow, newNumber);
    }
    bool clearNumber();
    void revertNumber();

    const Geometry& getGeometry() const { return m_geometry; }

    unsigned getMaxNumber() const {
      return getGeometry().getMaxNumber();
    }

    unsigned getRowIndex() const {
      return getGeometry().getRowIndex(getSquareIndex());
    }
    unsigned getColumnIndex() const {
      return getGeometry().getColumnIndex(getSquareIndex());
    }

    unsigned getTag() const { return m_tag; }
    void setTag(unsigned n) { m_tag = n; }
    void setTag(unsigned n) const { m_tag = n; }

    const Sequence* getRow() const { return m_row; }
    Sequence* getRow() { return m_row; }

    const Sequence* getColumn() const { return m_column; }
    Sequence* getColumn() { return m_column; }

    const Sequence* getGrid() const { return m_grid; }
    Sequence* getGrid() { return m_grid; }

    unsigned getPossibilities() const { return m_state.possibilities; }
    bitmask_t getForbidden() const { return m_state.forbidden; }

    bool forbid_nothrow(unsigned number);
    bool forbid(unsigned number);
    bool forbid_nothrow(const Bitmask&);
    bool forbid(const Bitmask&);
    bool allow(unsigned number);

    bool belongsTo(const Sequence*) const;

    void dropStack() { m_stack.clear(); }
    void clear();

    void push();
    bool pop();

    Square(const Geometry&, unsigned i);
    ~Square();

    Square& operator=(unsigned n) {
        setNumber(n);
        return *this;
    }

    Square& operator=(const FIXED &f) {
        setNumber(f.getNumber());
        forbidAll();
        return *this;
    }

    friend class Board;
  };


  class FewestPossibilitiesFirst
  {
  public:
    bool operator()(const Square *s1, const Square *s2) const {
      return (s1->getPossibilities() < s2->getPossibilities());
    }
  };
}

#endif
