#include <iostream>
using namespace std;

#include "strategy.h"
#include "freedom.h"
#include "puzzle.h"
#include "sequence.h"
#include "square.h"

namespace sudoku
{

Strategy Strategy::ApplyAll(checkAll);
Strategy Strategy::FindCompleteSequences(checkCompleteSequences);
Strategy Strategy::FindCrossSequenceConstraints(checkCrossSequenceConstraints);
Strategy Strategy::FindNumberSubsets(checkNumberSubsets);


/*********************************************************************
 */
int
Strategy::checkAll(Puzzle *puzzle, const Freedom &freedom)
{
  int changes = 0;

  const StrategyCollection::const_iterator end = puzzle->getStrategies().end();
  StrategyCollection::const_iterator strategy = puzzle->getStrategies().begin();

  while (strategy != end)
  {
    changes = (*strategy)(puzzle, freedom);
    if (changes != 0)
    {
      break;
    }

    ++strategy;
  } /* while */

  return changes;
}


/*********************************************************************
 */
int
Strategy::checkCompleteSequences(Puzzle *puzzle, const Freedom &freedom)
{
  TRACE( cout << "COMPLETING SEQUENCES" << endl; )

  Sequence *sequence = freedom.getSequence();
  int changes = 0;

  for (Freedom::size_type n = 1; n < freedom.size(); ++n)
  {
    if (freedom[n].getCount() != 0)
    {
      TRACE( cout << "- put " << n
                  << " in " << freedom[n]
                  << endl; )

      if (freedom[n].getCount() == 1)
      {
        const unsigned sqIndex = freedom[n].getFirstBit();
        Square *square = &(*sequence)[sqIndex];

        if (square->getNumber() == 0)
        {
          puzzle->addCompletedSequence(square, sequence, n);
          square->setNumber(n);
          ++changes;

          TRACE( cout << "Number " << n
                      << " must be placed in square " << freedom[n]
                      << endl; )
        }
      }
    }
  } /* for */

  TRACE( cout << "DONE: changes=" << changes << endl; )

  return changes;
}


/*********************************************************************
 */
int
Strategy::checkCrossSequenceConstraints(Puzzle *puzzle,
                                        const Freedom &freedom)
{
  TRACE( cout << "CHECKING FOR CROSS-SEQUENCE CONSTRAINTS" << endl; )

  const Sequence *sequence = freedom.getSequence();
  int changes = 0;

  for (Freedom::size_type n = 1; n < freedom.size(); ++n)
  {
    Sequence *cross = sequence->findCrossSequence(freedom[n]);
    if ((cross != NULL) && cross->isMissing(n))
    {
      const SquareSequence::iterator endSquare = cross->getSquares().end();
      SquareSequence::iterator sqiter = cross->getSquares().begin();

      while (sqiter != endSquare)
      {
        Square *square = *sqiter;

        try
        {
          if ( (square->getNumber() == 0) &&
               !square->belongsTo(sequence) &&
               square->forbid(n) )
          {
            puzzle->addCrossSequenceConstraint(square, sequence, cross, n);
            ++changes;
          }
        }
        catch(...)
        {
          puzzle->addCrossSequenceConstraint(square, sequence, cross, n);
          throw;
        }

        ++sqiter;
      } /* while */
    }
  } /* for */

  return changes;
}


/*********************************************************************
 */
int
Strategy::checkNumberSubsets(Puzzle *puzzle, const Freedom &freedom)
{
  TRACE( cout << "CHECKING FOR SUBSETS OF NUMBERS" << endl; )

  int totalChanges = 0;

  SubsetCollection subsets;
  if ( freedom.findNumberSubsets(subsets) )
  {
    const SubsetCollection::const_iterator end = subsets.end();
    SubsetCollection::const_iterator subset = subsets.begin();

    Sequence *sequence = freedom.getSequence();
    const bitmask_t numberMask = sequence->getNumberMask();

    while (subset != end)
    {
      const Bitmask inverse( ~subset->getNumbers().getBits() & numberMask );
      int changes = 0;

      try
      {
        const Sequence::size_type size = sequence->size();
        for (Sequence::size_type i = 0; i < size; ++i)
        {
          if ( subset->getSquares().isSet(i) )
          {
            Square *square = &(*sequence)[i];

            TRACE( cout << "Forbidding " << square->getName()
                        << " to be " << inverse
                        << endl; )

            if ( (square->getNumber() == 0) &&
                 (square->forbid(inverse)) )
            {
              ++changes;
            }
          }
        } /* for */

        if (changes != 0)
        {
          puzzle->addNumberSubset(sequence, *subset);
          totalChanges += changes;
        }
      }
      catch(...)
      {
        puzzle->addNumberSubset(sequence, *subset);
        throw;
      }

      ++subset;
    } /* while */
  }

  return totalChanges;
}

}

